package org.greenstone.android.tipple.base;

import java.text.DecimalFormat;
import java.util.Collections;
import java.util.Comparator;

import org.mapsforge.core.GeoPoint;

import android.app.AlertDialog;
import android.content.DialogInterface;
import android.content.DialogInterface.OnCancelListener;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.location.Location;
import android.speech.tts.TextToSpeech;
import android.view.MotionEvent;
import android.view.View;

public class AROverlay extends View {
	ARActivity activity;
	Paint paintFill;
	Paint paintOutline;
	Paint paintText;
	Paint borderText;
	
	DecimalFormat df;

	boolean isDialogBoxShowing = false;

	public AROverlay(ARActivity context) {
		super(context);
		
		df = new DecimalFormat("#.#");

		activity = context;
		paintFill = new Paint();
		paintFill.setStyle(Paint.Style.FILL);
		paintFill.setAntiAlias(true);
		paintOutline = new Paint();
		paintOutline.setStyle(Paint.Style.STROKE);
		paintOutline.setStrokeWidth(5);
		paintOutline.setAntiAlias(true);
		paintText = new Paint();
		paintText.setTextAlign(Paint.Align.CENTER);
		paintText.setTextSize(35);
		paintText.setColor(Color.BLACK);
		paintText.setAntiAlias(true);
		paintText.setShadowLayer(3, 1, 1, Color.WHITE);
	}

	@Override
	public boolean onTouchEvent(MotionEvent event) {
		
		if (isDialogBoxShowing) return false;
		
		float x_finger = event.getX();
		float y_finger = event.getY();

		for (int i = Global.locations.size() - 1; i >= 0; i--)
		{
			TipLocation loc = Global.locations.get(i);
			
			if (loc.hits(x_finger, y_finger))
			{
				isDialogBoxShowing = true;
				
				AlertDialog.Builder dialog = new AlertDialog.Builder(activity);
				dialog.setTitle(loc.getTitle());
				dialog.setMessage(loc.getText());
				
				dialog.setOnCancelListener(new OnCancelListener() {
					@Override
					public void onCancel(DialogInterface dialog) {
						Global.tts.stop();
						isDialogBoxShowing = false;
					}
				});
				
				dialog.show();
				
				Global.tts.speak(loc.getTitleTTS(), TextToSpeech.QUEUE_FLUSH, null);
				Global.tts.speak(loc.getTTSText(), TextToSpeech.QUEUE_ADD, null);
				
				ArDirection.setGPSpidouble(new GeoPoint(loc.getLatitude(), loc.getLongitude()));
				return true;
			}
		}
		
		return false;
	}

	@Override
	protected void onDraw(Canvas canvas) {

		// recalculate the screen locations
		for (int i = 0; i < Global.locations.size(); i++) {
			Global.locations.get(i).projectToScreen(LocationFusion.getInstance().lastKnownLocation,
					SensorFusion.getInstance().getRotationMatrix());
		}

		// sort the points, closest first
		Collections.sort(Global.locations, new Comparator<TipLocation>() {
			public int compare(TipLocation left, TipLocation right) {
				return Float.compare(left.getDistance(), right.getDistance());
			}
		});

		float screenAngle = SensorFusion.getInstance().getScreenAngleDegrees();
		float[] markerHitCircle = new float[3];

		// draw the locations (closest last)
		for (int i = Global.locations.size() - 1; i >= 0; i--) {

			TipLocation loc = Global.locations.get(i);

			if (loc.isVisible()) {

				canvas.save(); // save the default canvas matrix

				// shift the canvas so that the location can be drawn at (0, 0)
				canvas.translate(loc.getScreenX(), loc.getScreenY());
				// rotate the canvas
				canvas.rotate(screenAngle);
				
				canvas.save(); // save the rotated & translated canvas

				// scale the canvas for the text
				float textScale = Math.min(200 * loc.getScreenScale(), 1);
				canvas.scale(textScale, textScale);
				
				// draw the text
				//String text = String.format("%s (%1.1fm)", loc.getTitle(), loc.getDistance());
				canvas.drawText(loc.getTitle() + " (" + df.format(loc.getDistance()) + "m)", 0, paintText.getTextSize(), paintText);
				
				canvas.restore(); // undo the scaling (go back to the rotated & scaled version)

				// scale the canvas for the bitmap
				float markerScale = Math.min(500 * loc.getScreenScale(), 4);
				canvas.scale(markerScale, markerScale);

				// now draw the bitmap onto the centre of the canvas and let the canvas' matrix
				// transform that to the right location, rotation and size
				float x = -loc.getMarkerBitmap().getWidth() / 2;
				float y = -loc.getMarkerBitmap().getHeight();
				canvas.drawBitmap(loc.getMarkerBitmap(), x, y, paintFill);
				
				// save the marker hit circle
				markerHitCircle[0] = 0;
				markerHitCircle[1] = -loc.getMarkerBitmap().getHeight() / 2;
				canvas.getMatrix().mapPoints(markerHitCircle);
				loc.setHitCircle(markerHitCircle[0], markerHitCircle[1], markerScale * 12 + 5);

				canvas.restore(); // undo the rotation & translation
			}
		}
		//drawInfo(canvas);

		super.onDraw(canvas);
	}

	private void drawInfo(Canvas canvas) {
		Paint back = new Paint();
		back.setColor(Color.WHITE);
		back.setAlpha(150);
		Paint fore = new Paint();
		fore.setColor(Color.BLACK);
		fore.setTextSize(25);
		fore.setAntiAlias(true);

		Location loc = LocationFusion.getInstance().lastKnownLocation;
		String[] text = new String[] {
				String.format("GPS: lat=%.1f, lon=%.1f, alt=%.1f, accuracy=%.2f",
						loc.getLatitude(), loc.getLongitude(), loc.getAltitude(), loc.getAccuracy()),
				String.format("Number of points: %d", Global.locations.size()) };

		canvas.drawRect(0, 0, 800, 100, back);

		for (int row = 0; row < text.length; row++) {
			canvas.drawText(text[row], 10, 30 + 30 * row, fore);
		}
	}
}
